import asyncio
import json
import platform

import pytest
from tornado.httpclient import HTTPClientError
from traitlets.config import Config


@pytest.fixture(params=["MappingKernelManager", "AsyncMappingKernelManager"])
def jp_argv(request):
    return [
        "--ServerApp.kernel_manager_class=jupyter_server.services.kernels.kernelmanager."
        + request.param
    ]


CULL_TIMEOUT = 30 if platform.python_implementation() == "PyPy" else 5
CULL_INTERVAL = 1


@pytest.fixture
def jp_server_config():
    return Config(
        {
            "ServerApp": {
                "MappingKernelManager": {
                    "cull_idle_timeout": CULL_TIMEOUT,
                    "cull_interval": CULL_INTERVAL,
                    "cull_connected": False,
                }
            }
        }
    )


async def test_culling(jp_fetch, jp_ws_fetch, jp_cleanup_subprocesses):
    r = await jp_fetch("api", "kernels", method="POST", allow_nonstandard_methods=True)
    kernel = json.loads(r.body.decode())
    kid = kernel["id"]

    # Open a websocket connection.
    ws = await jp_ws_fetch("api", "kernels", kid, "channels")

    r = await jp_fetch("api", "kernels", kid, method="GET")
    model = json.loads(r.body.decode())
    assert model["connections"] == 1
    culled = await get_cull_status(kid, jp_fetch)  # connected, should not be culled
    assert not culled
    ws.close()
    culled = await get_cull_status(kid, jp_fetch)  # not connected, should be culled
    assert culled
    await jp_cleanup_subprocesses()


async def get_cull_status(kid, jp_fetch):
    frequency = 0.5
    culled = False
    for _ in range(
        int((CULL_TIMEOUT + CULL_INTERVAL) / frequency)
    ):  # Timeout + Interval will ensure cull
        try:
            r = await jp_fetch("api", "kernels", kid, method="GET")
            json.loads(r.body.decode())
        except HTTPClientError as e:
            assert e.code == 404
            culled = True
            break
        else:
            await asyncio.sleep(frequency)
    return culled



"""Tornado handlers for api specifications."""
# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import json
import os

from tornado import web

from ...base.handlers import APIHandler
from ...base.handlers import JupyterHandler
from jupyter_server._tz import isoformat
from jupyter_server._tz import utcfromtimestamp
from jupyter_server.utils import ensure_async


class APISpecHandler(web.StaticFileHandler, JupyterHandler):
    def initialize(self):
        web.StaticFileHandler.initialize(self, path=os.path.dirname(__file__))

    @web.authenticated
    def get(self):
        self.log.warning("Serving api spec (experimental, incomplete)")
        return web.StaticFileHandler.get(self, "api.yaml")

    def get_content_type(self):
        return "text/x-yaml"


class APIStatusHandler(APIHandler):

    _track_activity = False

    @web.authenticated
    async def get(self):
        # if started was missing, use unix epoch
        started = self.settings.get("started", utcfromtimestamp(0))
        started = isoformat(started)

        kernels = await ensure_async(self.kernel_manager.list_kernels())
        total_connections = sum(k["connections"] for k in kernels)
        last_activity = isoformat(self.application.last_activity())
        model = {
            "started": started,
            "last_activity": last_activity,
            "kernels": len(kernels),
            "connections": total_connections,
        }
        self.finish(json.dumps(model, sort_keys=True))


default_handlers = [
    (r"/api/spec.yaml", APISpecHandler),
    (r"/api/status", APIStatusHandler),
]



import json


async def test_list_formats(jp_fetch):
    r = await jp_fetch("api", "nbconvert", method="GET")
    formats = json.loads(r.body.decode())
    # Verify the type of the response.
    assert isinstance(formats, dict)
    # Verify that all returned formats have an
    # output mimetype defined.
    required_keys_present = []
    for name, data in formats.items():
        required_keys_present.append("output_mimetype" in data)
    assert all(required_keys_present), "All returned formats must have a `output_mimetype` key."



# coding: utf-8
import json
from base64 import encodebytes
from shutil import which

import pytest
import tornado
from nbformat import writes
from nbformat.v4 import new_code_cell
from nbformat.v4 import new_markdown_cell
from nbformat.v4 import new_notebook
from nbformat.v4 import new_output

from ..utils import expected_http_error


png_green_pixel = encodebytes(
    b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00"
    b"\x00\x00\x01\x00\x00x00\x01\x08\x02\x00\x00\x00\x90wS\xde\x00\x00\x00\x0cIDAT"
    b"\x08\xd7c\x90\xfb\xcf\x00\x00\x02\\\x01\x1e.~d\x87\x00\x00\x00\x00IEND\xaeB`\x82"
).decode("ascii")


@pytest.fixture
def notebook(jp_root_dir):
    # Build sub directory.
    subdir = jp_root_dir / "foo"
    if not jp_root_dir.joinpath("foo").is_dir():
        subdir.mkdir()

    # Build a notebook programmatically.
    nb = new_notebook()
    nb.cells.append(new_markdown_cell(u"Created by test ³"))
    cc1 = new_code_cell(source=u"print(2*6)")
    cc1.outputs.append(new_output(output_type="stream", text=u"12"))
    cc1.outputs.append(
        new_output(
            output_type="execute_result",
            data={"image/png": png_green_pixel},
            execution_count=1,
        )
    )
    nb.cells.append(cc1)

    # Write file to tmp dir.
    nbfile = subdir / "testnb.ipynb"
    nbfile.write_text(writes(nb, version=4), encoding="utf-8")


pytestmark = pytest.mark.skipif(not which("pandoc"), reason="Command 'pandoc' is not available")


async def test_from_file(jp_fetch, notebook):
    r = await jp_fetch(
        "nbconvert", "html", "foo", "testnb.ipynb", method="GET", params={"download": False}
    )

    assert r.code == 200
    assert "text/html" in r.headers["Content-Type"]
    assert "Created by test" in r.body.decode()
    assert "print" in r.body.decode()

    r = await jp_fetch(
        "nbconvert", "python", "foo", "testnb.ipynb", method="GET", params={"download": False}
    )

    assert r.code == 200
    assert "text/x-python" in r.headers["Content-Type"]
    assert "print(2*6)" in r.body.decode()


async def test_from_file_404(jp_fetch, notebook):
    with pytest.raises(tornado.httpclient.HTTPClientError) as e:
        await jp_fetch(
            "nbconvert",
            "html",
            "foo",
            "thisdoesntexist.ipynb",
            method="GET",
            params={"download": False},
        )
    assert expected_http_error(e, 404)


async def test_from_file_download(jp_fetch, notebook):
    r = await jp_fetch(
        "nbconvert", "python", "foo", "testnb.ipynb", method="GET", params={"download": True}
    )
    content_disposition = r.headers["Content-Disposition"]
    assert "attachment" in content_disposition
    assert "testnb.py" in content_disposition


async def test_from_file_zip(jp_fetch, notebook):
    r = await jp_fetch(
        "nbconvert", "latex", "foo", "testnb.ipynb", method="GET", params={"download": True}
    )
    assert "application/zip" in r.headers["Content-Type"]
    assert ".zip" in r.headers["Content-Disposition"]


async def test_from_post(jp_fetch, notebook):
    r = await jp_fetch(
        "api/contents/foo/testnb.ipynb",
        method="GET",
    )
    nbmodel = json.loads(r.body.decode())

    r = await jp_fetch("nbconvert", "html", method="POST", body=json.dumps(nbmodel))
    assert r.code == 200
    assert "text/html" in r.headers["Content-Type"]
    assert "Created by test" in r.body.decode()
    assert "print" in r.body.decode()

    r = await jp_fetch("nbconvert", "python", method="POST", body=json.dumps(nbmodel))
    assert r.code == 200
    assert u"text/x-python" in r.headers["Content-Type"]
    assert "print(2*6)" in r.body.decode()


async def test_from_post_zip(jp_fetch, notebook):
    r = await jp_fetch(
        "api/contents/foo/testnb.ipynb",
        method="GET",
    )
    nbmodel = json.loads(r.body.decode())

    r = await jp_fetch("nbconvert", "latex", method="POST", body=json.dumps(nbmodel))
    assert "application/zip" in r.headers["Content-Type"]
    assert ".zip" in r.headers["Content-Disposition"]



import pytest

from jupyter_server.services.contents.checkpoints import AsyncCheckpoints
from jupyter_server.services.contents.filecheckpoints import AsyncGenericFileCheckpoints
from jupyter_server.services.contents.filecheckpoints import GenericFileCheckpoints
from jupyter_server.services.contents.manager import AsyncContentsManager


@pytest.fixture(params=[AsyncGenericFileCheckpoints, GenericFileCheckpoints])
def jp_server_config(request):
    return {"FileContentsManager": {"checkpoints_class": request.param}}


def test_config_did_something(jp_server_config, jp_serverapp):
    assert isinstance(
        jp_serverapp.contents_manager.checkpoints,
        jp_server_config["FileContentsManager"]["checkpoints_class"],
    )


async def test_async_contents_manager(jp_configurable_serverapp):
    config = {"ContentsManager": {"checkpoints_class": AsyncCheckpoints}}
    argv = [
        "--ServerApp.contents_manager_class=jupyter_server.services.contents.manager.AsyncContentsManager"
    ]
    app = jp_configurable_serverapp(config=config, argv=argv)
    assert isinstance(app.contents_manager, AsyncContentsManager)



"""Tornado handlers for security logging."""
# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
from tornado import web

from . import csp_report_uri
from ...base.handlers import APIHandler


class CSPReportHandler(APIHandler):
    """Accepts a content security policy violation report"""

    _track_activity = False

    def skip_check_origin(self):
        """Don't check origin when reporting origin-check violations!"""
        return True

    def check_xsrf_cookie(self):
        # don't check XSRF for CSP reports
        return

    @web.authenticated
    def post(self):
        """Log a content security policy violation report"""
        self.log.warning(
            "Content security violation: %s", self.request.body.decode("utf8", "replace")
        )


default_handlers = [(csp_report_uri, CSPReportHandler)]



# encoding: utf-8
"""Tornado handlers for the terminal emulator."""
# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import terminado
from tornado import web

from ..base.handlers import JupyterHandler
from ..base.zmqhandlers import WebSocketMixin
from jupyter_server._tz import utcnow


class TermSocket(WebSocketMixin, JupyterHandler, terminado.TermSocket):
    def origin_check(self):
        """Terminado adds redundant origin_check
        Tornado already calls check_origin, so don't do anything here.
        """
        return True

    def get(self, *args, **kwargs):
        if not self.get_current_user():
            raise web.HTTPError(403)
        if not args[0] in self.term_manager.terminals:
            raise web.HTTPError(404)
        return super(TermSocket, self).get(*args, **kwargs)

    def on_message(self, message):
        super(TermSocket, self).on_message(message)
        self._update_activity()

    def write_message(self, message, binary=False):
        super(TermSocket, self).write_message(message, binary=binary)
        self._update_activity()

    def _update_activity(self):
        self.application.settings["terminal_last_activity"] = utcnow()
        # terminal may not be around on deletion/cull
        if self.term_name in self.terminal_manager.terminals:
            self.terminal_manager.terminals[self.term_name].last_activity = utcnow()



from jupyter_server.base.handlers import JupyterHandler
from jupyter_server.extension.handler import ExtensionHandlerJinjaMixin
from jupyter_server.extension.handler import ExtensionHandlerMixin
from jupyter_server.utils import url_escape


class DefaultHandler(ExtensionHandlerMixin, JupyterHandler):
    def get(self):
        # The name of the extension to which this handler is linked.
        self.log.info("Extension Name in {} Default Handler: {}".format(self.name, self.name))
        # A method for getting the url to static files (prefixed with /static/<name>).
        self.log.info(
            "Static URL for / in simple_ext1 Default Handler: {}".format(self.static_url(path="/"))
        )
        self.write("<h1>Hello Simple 1 - I am the default...</h1>")
        self.write("Config in {} Default Handler: {}".format(self.name, self.config))


class RedirectHandler(ExtensionHandlerMixin, JupyterHandler):
    def get(self):
        self.redirect("/static/{}/favicon.ico".format(self.name))


class ParameterHandler(ExtensionHandlerMixin, JupyterHandler):
    def get(self, matched_part=None, *args, **kwargs):
        var1 = self.get_argument("var1", default=None)
        components = [x for x in self.request.path.split("/") if x]
        self.write("<h1>Hello Simple App 1 from Handler.</h1>")
        self.write("<p>matched_part: {}</p>".format(url_escape(matched_part)))
        self.write("<p>var1: {}</p>".format(url_escape(var1)))
        self.write("<p>components: {}</p>".format(components))


class BaseTemplateHandler(ExtensionHandlerJinjaMixin, ExtensionHandlerMixin, JupyterHandler):
    pass


class TypescriptHandler(BaseTemplateHandler):
    def get(self):
        self.write(self.render_template("typescript.html"))


class TemplateHandler(BaseTemplateHandler):
    def get(self, path):
        """ Optionaly, you can print(self.get_template('simple1.html'))"""
        self.write(self.render_template("simple1.html", path=path))


class ErrorHandler(BaseTemplateHandler):
    def get(self, path):
        self.write(self.render_template("error.html", path=path))



"""
Classes for managing Checkpoints.
"""
# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
from tornado.web import HTTPError
from traitlets.config.configurable import LoggingConfigurable


class Checkpoints(LoggingConfigurable):
    """
    Base class for managing checkpoints for a ContentsManager.

    Subclasses are required to implement:

    create_checkpoint(self, contents_mgr, path)
    restore_checkpoint(self, contents_mgr, checkpoint_id, path)
    rename_checkpoint(self, checkpoint_id, old_path, new_path)
    delete_checkpoint(self, checkpoint_id, path)
    list_checkpoints(self, path)
    """

    def create_checkpoint(self, contents_mgr, path):
        """Create a checkpoint."""
        raise NotImplementedError("must be implemented in a subclass")

    def restore_checkpoint(self, contents_mgr, checkpoint_id, path):
        """Restore a checkpoint"""
        raise NotImplementedError("must be implemented in a subclass")

    def rename_checkpoint(self, checkpoint_id, old_path, new_path):
        """Rename a single checkpoint from old_path to new_path."""
        raise NotImplementedError("must be implemented in a subclass")

    def delete_checkpoint(self, checkpoint_id, path):
        """delete a checkpoint for a file"""
        raise NotImplementedError("must be implemented in a subclass")

    def list_checkpoints(self, path):
        """Return a list of checkpoints for a given file"""
        raise NotImplementedError("must be implemented in a subclass")

    def rename_all_checkpoints(self, old_path, new_path):
        """Rename all checkpoints for old_path to new_path."""
        for cp in self.list_checkpoints(old_path):
            self.rename_checkpoint(cp["id"], old_path, new_path)

    def delete_all_checkpoints(self, path):
        """Delete all checkpoints for the given path."""
        for checkpoint in self.list_checkpoints(path):
            self.delete_checkpoint(checkpoint["id"], path)


class GenericCheckpointsMixin(object):
    """
    Helper for creating Checkpoints subclasses that can be used with any
    ContentsManager.

    Provides a ContentsManager-agnostic implementation of `create_checkpoint`
    and `restore_checkpoint` in terms of the following operations:

    - create_file_checkpoint(self, content, format, path)
    - create_notebook_checkpoint(self, nb, path)
    - get_file_checkpoint(self, checkpoint_id, path)
    - get_notebook_checkpoint(self, checkpoint_id, path)

    To create a generic CheckpointManager, add this mixin to a class that
    implement the above four methods plus the remaining Checkpoints API
    methods:

    - delete_checkpoint(self, checkpoint_id, path)
    - list_checkpoints(self, path)
    - rename_checkpoint(self, checkpoint_id, old_path, new_path)
    """

    def create_checkpoint(self, contents_mgr, path):
        model = contents_mgr.get(path, content=True)
        type = model["type"]
        if type == "notebook":
            return self.create_notebook_checkpoint(
                model["content"],
                path,
            )
        elif type == "file":
            return self.create_file_checkpoint(
                model["content"],
                model["format"],
                path,
            )
        else:
            raise HTTPError(500, u"Unexpected type %s" % type)

    def restore_checkpoint(self, contents_mgr, checkpoint_id, path):
        """Restore a checkpoint."""
        type = contents_mgr.get(path, content=False)["type"]
        if type == "notebook":
            model = self.get_notebook_checkpoint(checkpoint_id, path)
        elif type == "file":
            model = self.get_file_checkpoint(checkpoint_id, path)
        else:
            raise HTTPError(500, u"Unexpected type %s" % type)
        contents_mgr.save(model, path)

    # Required Methods
    def create_file_checkpoint(self, content, format, path):
        """Create a checkpoint of the current state of a file

        Returns a checkpoint model for the new checkpoint.
        """
        raise NotImplementedError("must be implemented in a subclass")

    def create_notebook_checkpoint(self, nb, path):
        """Create a checkpoint of the current state of a file

        Returns a checkpoint model for the new checkpoint.
        """
        raise NotImplementedError("must be implemented in a subclass")

    def get_file_checkpoint(self, checkpoint_id, path):
        """Get the content of a checkpoint for a non-notebook file.

        Returns a dict of the form:
        {
            'type': 'file',
            'content': <str>,
            'format': {'text','base64'},
        }
        """
        raise NotImplementedError("must be implemented in a subclass")

    def get_notebook_checkpoint(self, checkpoint_id, path):
        """Get the content of a checkpoint for a notebook.

        Returns a dict of the form:
        {
            'type': 'notebook',
            'content': <output of nbformat.read>,
        }
        """
        raise NotImplementedError("must be implemented in a subclass")


class AsyncCheckpoints(Checkpoints):
    """
    Base class for managing checkpoints for a ContentsManager asynchronously.
    """

    async def create_checkpoint(self, contents_mgr, path):
        """Create a checkpoint."""
        raise NotImplementedError("must be implemented in a subclass")

    async def restore_checkpoint(self, contents_mgr, checkpoint_id, path):
        """Restore a checkpoint"""
        raise NotImplementedError("must be implemented in a subclass")

    async def rename_checkpoint(self, checkpoint_id, old_path, new_path):
        """Rename a single checkpoint from old_path to new_path."""
        raise NotImplementedError("must be implemented in a subclass")

    async def delete_checkpoint(self, checkpoint_id, path):
        """delete a checkpoint for a file"""
        raise NotImplementedError("must be implemented in a subclass")

    async def list_checkpoints(self, path):
        """Return a list of checkpoints for a given file"""
        raise NotImplementedError("must be implemented in a subclass")

    async def rename_all_checkpoints(self, old_path, new_path):
        """Rename all checkpoints for old_path to new_path."""
        for cp in await self.list_checkpoints(old_path):
            await self.rename_checkpoint(cp["id"], old_path, new_path)

    async def delete_all_checkpoints(self, path):
        """Delete all checkpoints for the given path."""
        for checkpoint in await self.list_checkpoints(path):
            await self.delete_checkpoint(checkpoint["id"], path)


class AsyncGenericCheckpointsMixin(GenericCheckpointsMixin):
    """
    Helper for creating Asynchronous Checkpoints subclasses that can be used with any
    ContentsManager.
    """

    async def create_checkpoint(self, contents_mgr, path):
        model = await contents_mgr.get(path, content=True)
        type = model["type"]
        if type == "notebook":
            return await self.create_notebook_checkpoint(
                model["content"],
                path,
            )
        elif type == "file":
            return await self.create_file_checkpoint(
                model["content"],
                model["format"],
                path,
            )
        else:
            raise HTTPError(500, u"Unexpected type %s" % type)

    async def restore_checkpoint(self, contents_mgr, checkpoint_id, path):
        """Restore a checkpoint."""
        type = await contents_mgr.get(path, content=False)["type"]
        if type == "notebook":
            model = await self.get_notebook_checkpoint(checkpoint_id, path)
        elif type == "file":
            model = await self.get_file_checkpoint(checkpoint_id, path)
        else:
            raise HTTPError(500, u"Unexpected type %s" % type)
        await contents_mgr.save(model, path)

    # Required Methods
    async def create_file_checkpoint(self, content, format, path):
        """Create a checkpoint of the current state of a file

        Returns a checkpoint model for the new checkpoint.
        """
        raise NotImplementedError("must be implemented in a subclass")

    async def create_notebook_checkpoint(self, nb, path):
        """Create a checkpoint of the current state of a file

        Returns a checkpoint model for the new checkpoint.
        """
        raise NotImplementedError("must be implemented in a subclass")

    async def get_file_checkpoint(self, checkpoint_id, path):
        """Get the content of a checkpoint for a non-notebook file.

        Returns a dict of the form:
        {
            'type': 'file',
            'content': <str>,
            'format': {'text','base64'},
        }
        """
        raise NotImplementedError("must be implemented in a subclass")

    async def get_notebook_checkpoint(self, checkpoint_id, path):
        """Get the content of a checkpoint for a notebook.

        Returns a dict of the form:
        {
            'type': 'notebook',
            'content': <output of nbformat.read>,
        }
        """
        raise NotImplementedError("must be implemented in a subclass")



import os

from traitlets import Unicode

from .handlers import DefaultHandler
from .handlers import ErrorHandler
from .handlers import ParameterHandler
from .handlers import RedirectHandler
from .handlers import TemplateHandler
from .handlers import TypescriptHandler
from jupyter_server.extension.application import ExtensionApp
from jupyter_server.extension.application import ExtensionAppJinjaMixin

DEFAULT_STATIC_FILES_PATH = os.path.join(os.path.dirname(__file__), "static")
DEFAULT_TEMPLATE_FILES_PATH = os.path.join(os.path.dirname(__file__), "templates")


class SimpleApp1(ExtensionAppJinjaMixin, ExtensionApp):

    # The name of the extension.
    name = "simple_ext1"

    # The url that your extension will serve its homepage.
    extension_url = "/simple_ext1/default"

    # Should your extension expose other server extensions when launched directly?
    load_other_extensions = True

    # Local path to static files directory.
    static_paths = [DEFAULT_STATIC_FILES_PATH]

    # Local path to templates directory.
    template_paths = [DEFAULT_TEMPLATE_FILES_PATH]

    configA = Unicode("", config=True, help="Config A example.")

    configB = Unicode("", config=True, help="Config B example.")

    configC = Unicode("", config=True, help="Config C example.")

    def initialize_handlers(self):
        self.handlers.extend(
            [
                (r"/{}/default".format(self.name), DefaultHandler),
                (r"/{}/params/(.+)$".format(self.name), ParameterHandler),
                (r"/{}/template1/(.*)$".format(self.name), TemplateHandler),
                (r"/{}/redirect".format(self.name), RedirectHandler),
                (r"/{}/typescript/?".format(self.name), TypescriptHandler),
                (r"/{}/(.*)", ErrorHandler),
            ]
        )

    def initialize_settings(self):
        self.log.info("Config {}".format(self.config))


# -----------------------------------------------------------------------------
# Main entry point
# -----------------------------------------------------------------------------

main = launch_new_instance = SimpleApp1.launch_instance



import os
import sys

import terminado

from ..utils import check_version

if not check_version(terminado.__version__, "0.8.3"):
    raise ImportError("terminado >= 0.8.3 required, found %s" % terminado.__version__)

from ipython_genutils.py3compat import which
from jupyter_server.utils import url_path_join as ujoin
from . import api_handlers
from .handlers import TermSocket
from .terminalmanager import TerminalManager


def initialize(webapp, root_dir, connection_url, settings):
    if os.name == "nt":
        default_shell = "powershell.exe"
    else:
        default_shell = which("sh")
    shell_override = settings.get("shell_command")
    shell = [os.environ.get("SHELL") or default_shell] if shell_override is None else shell_override
    # When the notebook server is not running in a terminal (e.g. when
    # it's launched by a JupyterHub spawner), it's likely that the user
    # environment hasn't been fully set up. In that case, run a login
    # shell to automatically source /etc/profile and the like, unless
    # the user has specifically set a preferred shell command.
    if os.name != "nt" and shell_override is None and not sys.stdout.isatty():
        shell.append("-l")
    terminal_manager = webapp.settings["terminal_manager"] = TerminalManager(
        shell_command=shell,
        extra_env={
            "JUPYTER_SERVER_ROOT": root_dir,
            "JUPYTER_SERVER_URL": connection_url,
        },
        parent=webapp.settings["serverapp"],
    )
    terminal_manager.log = webapp.settings["serverapp"].log
    base_url = webapp.settings["base_url"]
    handlers = [
        (
            ujoin(base_url, r"/terminals/websocket/(\w+)"),
            TermSocket,
            {"term_manager": terminal_manager},
        ),
        (ujoin(base_url, r"/api/terminals"), api_handlers.TerminalRootHandler),
        (ujoin(base_url, r"/api/terminals/(\w+)"), api_handlers.TerminalHandler),
    ]
    webapp.add_handlers(".*$", handlers)
